""" domible/src/domible/builders/navBuilder.py

here we have a class to build a navigation element, <nav>.
It's based on an article from web.dev:
https://web.dev/website-navigation/

I'm starting this on 2023-05-25, with a focus on bare minimum functionality.
There is likely a lot of static stuff, especially CSS and JS.

Ugh.  Now it's 2023-07-12 and I've made almost no coding progress.
I'm hung up on feature creep.
I've been reading different articles on how to implement navigation and support collapsible sub navigation and support arrow keys and escape and whatnot...
I've decided to initially support the most basic navigation, list of anchors.
And whatever the API looks like intially is almost certainly going to change.
I'm embracing the ideals of get something out there and iterate 
(maybe, or maybe not known as Minimum Viable Product)

the following is from https://developer.mozilla.org/en-US/docs/web/html/element/nav
  The <nav> element represents a section of a page whose purpose is to provide navigation links,
  either within the current document or to other documents. 
  Common examples of navigation sections are 
  menus, tables of contents, and indexes.
""" 

from typing import Any

from domible.elements import Nav
from domible.elements import UnorderedList, OrderedList 

class NavBuilder:
    """
    This minimalist navigation element consists of a <nav> element and list of navigable items.
    For screen reader support, we'll require a string label for the nav element
    Though trying to not hard code any assumptions in the API for this class, 
    my expectation is the list of navigable items will be anchors.
    """

    def __init__(self, label: str, items: list[Any] = None, ordered: bool = False, currentPage:int = 0, **attributes):
        """
        label will be used in an aria-label attribute for the nav element.
        items is the list of navigable links in the ul or ol list
        the value of ordered dictates whether a ul or ol list is used.
        currentPage is the index of the item that is the page currently accessed from the nav.
        currentPage is used to set the aria-current="page" attribute on the list item 
        """
        self.label = label
        self.items = items if items else [] 
        if not ordered:
            self.itemsList = UnorderedList(self.items)
        else:
            self.itemsList = OrderedList(self.items)

## end of file 


""" domible/src/domible/builders/elementFromObject.py 

Create fairly simple HTML from a Python object.

Okay, it's been many days since I wrote that opening comment.
I hadn't thought it through and now realize "simple" is not the right word.
But it's really not that bad, and I've written and deleted a lot of crap in this module comment.
So here it is at a high level.
anything not a list, tuple, set, or dict is considered a terminal.
lists, tuples, and sets are converted to simple unordered lists.
dicts are unordered lists with listitems as key: value pairs.
When a value in a dict, or an item in a list is a list, tuple, set, or dict, a sublist is created.
Sublists are presented as collapsible items.
See comments in the code for more details. 
"""

import logging 
logger = logging.getLogger(__name__)

import html 
import json 

from domible.elements import BaseElement, ListItem, UnorderedList, Paragraph, Anchor, Heading, Details, Summary  

import validators 

##
# using a global for depth to use depth to calculate heading levels
max_depth = None

def process_iter(obj: object, depth: int) -> UnorderedList:
    """
    create an HTML <ul> from the iterator  
    For each item in the iterator, 
    get the HTML for that item,
    wrap it in a ListItem,
    and add it to the list.

    if the item is another list/tuple/set, or a dict,
    a sublist is added to the list in a collapsible widget.
    meta-data is used as the summary for the sublist.
    """
    logger.debug(f"process_iter with object {obj}, depth {depth}")
    ul = UnorderedList(**{"aria-label":f"entering level  {depth}", "style": "list-style-type: none;"})
    for item in obj:
        list_contents = process_terminal(item)
        if depth < max_depth and isinstance(item, (list, tuple, set, dict)) and len(item) > 0:
            # if the itemis a non-empty iterable or dict, and we have not hit depth yet,
            # create a collapsible widget with the item meta-data as the Summary,
            # and the item as a list in the collapsible content.
            summary = Summary(list_contents)
            hidden_html = None 
            if isinstance(item, dict):
                hidden_html = process_dict(item, depth + 1)
            else:  # must be a list, tuple, or set 
                hidden_html = process_iter(item, depth + 1)
            list_contents = Details(summary, hidden_html)
        ul.add_item(ListItem(list_contents))
    return ul


def process_dict(obj: dict, depth: int) -> BaseElement:
    """
    See comments at the file level regarding why dicts are treated the way they are here.
    As you'll see below, keys in the dict are treated as terminals (as I've defined in this file)
    Yes, a tuple can be a key in a dict.
    I'm making a simplifying decision here (already spent way too much time on this module).
    If treating tuples as terminals here becomes a problem, maybe I'll change it.

    if the value results in a sublist,
    add a string after the key indicating the type and length of that value
    I think it adds clarity when using a screen reader 
    Added a disclosure widget to be able to collapse/expand sublist 
    """
    logger.debug(f"process_dict with object {obj}, depth {depth}")
    ul = UnorderedList(**{"aria-label": f"entering level {depth}", "style": "list-style-type: none;",})
    for key, value in obj.items():
        # key, value are handled based on the type of the value.
        # always treat the key as a "terminal" though.
        # and start by assuming the value is also a terminal and change if necessary
        list_contents = f"{process_terminal(key)}: {process_terminal(value)}"
        if depth < max_depth and isinstance(value, (list, tuple, set, dict)) and len(value) > 0:
            # if the value is a non-empty iterable or dict, and we have not hit depth yet,
            # create a collapsible widget with the key as the Summary,
            # and the value as a list in the collapsible content.
            summary = Summary(list_contents)
            hidden_html = None 
            if isinstance(value, dict):
                hidden_html = process_dict(value, depth + 1)
            else: # value is a list, set, or tuple
                hidden_html = process_iter(value, depth + 1)
            # now create the collapsible widget
            list_contents = Details(summary, hidden_html)
        ul.add_item(ListItem(list_contents))
    return ul


def process_terminal(obj) -> str:
    """
    process_terminal is called either when there is a terminal object 
    (not a dict, list, tuple, or set)
    or when depth has been reached and we don't want to go any deeper into the object 
    """
    if isinstance(obj, (dict, list, tuple, set) ): 
        if len(obj) == 0:
            return f"type empty  {type(obj).__name__}"
        else:
            return f"type {type(obj).__name__}"
    return html.escape(str(obj))


def element_from_object(obj: object, depth: int = 42) -> BaseElement: 
    """
    since, in theory, objects can have unlimited complexity, 
    depth can be specified to limit the amount of HTML returned.
    Once depth is reached, any non-terminal objects
    (e.g., dict, list, tuple, set) will be a string with some meta-data about the object.

    initially support a subset of Python objects:
    int, float, str, bool, list, tuple, set, and dict 
    anything else will have its str representation returned in a paragraph element
    For converting to HTML, list, tuple, and set can all be represented in an unordered list,
    and all three objects are iterators, thus can be processed the same way.
    int, float, str, and bool will be string representatives of themselves, 
    thus can be processed the same way as object types not supported here.

    read the file comments regarding how dicts are represented,
    it got more complicated than I had imagined.
    """
    global max_depth 
    max_depth = depth 
    if isinstance(obj, dict):
        return process_dict(obj, 0)
    elif isinstance(obj, (list, tuple, set)):
        return process_iter(obj, 0)
    return Paragraph(process_terminal(obj))


## end of file



""" domible/src/domible/__init__.py 
"""

# read version from installed package
from importlib.metadata import version
__version__ = version("domible")

import domible.elements
import domible.builders
import domible.starterDocuments 
from domible.tools import open_object_in_browser, open_html_in_browser, save_to_file 
## end of file 


""" domible/src/dicli/__init__.py 
"""

# read version from installed package
# using same version as domible as it's the main package 
from importlib.metadata import version
__version__ = version("domible")


from dicli import main 

def run():
    main.run()


## end of file



#!/usr/bin/env python
""" domible/src/dicli/main.py 
simple command line tool to test/experiment with domible 
"""

import logging
import jsonloggeriso8601datetime as jlidt

jlidt.setConfig()
logger = logging.getLogger(__name__)


import dicli.mdnElements as mdnElements

from domible.elements import Html, Head, Body, Title, Base
from domible.elements import (
    UnorderedList,
    OrderedList,
    DescriptionList,
    ListItem,
    DListItem,
    DescriptionTerm,
    DescriptionDef,
)
from domible.elements import Heading, Anchor, Paragraph
from domible.builders.tableBuilder import TableBuilder, build_table_from_dicts
from domible.starterDocuments import basic_head_empty_body
from domible.tools import open_html_in_browser

import argparse

global_parser = argparse.ArgumentParser(
    prog="dicli",
    description="command line utility for domible testing",
    formatter_class=argparse.RawTextHelpFormatter,
    epilog="Cheers!",
)


def simple(args) -> None:
    """create a very simple HTML document and open it in the default browser"""
    title = "BareBones with a Minimal Body"
    htmlDoc = basic_head_empty_body(title)
    body = htmlDoc.get_body_element()
    body.add_content(
        [
            Heading(1, title),
            Paragraph("Really, this is all you're going to get."),
            Paragraph(
                f'You might find more interesting content at {Anchor(href="https://blindgumption.com", contents="the Blind Gumption Website")}'
            ),
        ]
    )
    open_html_in_browser(htmlDoc)


def elements(args) -> None:
    """
    elements is used to test, and provide an example of, Table and the tableBuilder
    it will scrape HTML element reference info from MDN and present it in a table in your default browser
    The HTML is also saved to a passed in file, not saved if no file specified.
    """
    mdn_base_url: str = args.url_mdn
    lang: str = args.lang
    outputfile: str = args.outfile

    if outputfile:
        print(f"saving html output to file: {outputfile}")
    title = "Tables of HTML Elements Scraped from MDN "
    htmlDoc = basic_head_empty_body(title, lang)
    head = htmlDoc.get_head_element()
    head.add_content(Base(href=mdn_base_url))
    body = htmlDoc.get_body_element()
    (currentElementsTable, deprecatedElementsTable) = mdnElements.getElementsTables(
        mdn_base_url, lang
    )
    if not currentElementsTable or not deprecatedElementsTable:
        body.add_content(
            Heading(1, f"failed to scrape elements from {mdnElements.MdnAnchor}")
        )
    else:
        # building up the body of the html document
        currentTable, _, _ = currentElementsTable.get_table()
        deprecatedTable, _, _ = deprecatedElementsTable.get_table()
        body.add_content(
            [
                Heading(1, title),
                Paragraph(
                    f"Information in the below tables was scraped from {mdnElements.MdnAnchor}."
                ),
                Heading(2, "Currently Supported HTML Elements"),
                currentTable,
                Heading(2, "Deprecated HTML Elements"),
                deprecatedTable,
            ]
        )
    open_html_in_browser(htmlDoc)


def ctfd(args) -> None:
    """
    to test the buildTableFromDicts function from tableBuilder
    """
    lower: int = args.lower
    upper: int = args.upper
    rows = []
    for x in range(lower, upper + 1):
        row = {"base": x}
        row.update(dict([(y, x**y) for y in range(lower, upper + 1)]))
        rows.append(row)
    ## we have a matrix, now display it in an HTML table
    table = build_table_from_dicts("base number raised to column heading number", rows)
    title = f"Testing createTableFromDicts function in tableBuilder, lower is {lower}, upper is {upper}"
    htmlDoc = basic_head_empty_body(title)
    body = htmlDoc.get_body_element()
    body.add_content(
        [
            Heading(1, title),
            Heading(2, "Raise base to power (column heading number)"),
            table,
        ]
    )
    open_html_in_browser(htmlDoc)


def lists(args) -> None:
    """
    Testing/example of using the lists related elements
    Note how elements are created using composition

    we're testing three types of lists, so let's make some coffee:
    unordered (ul) - stuff we need to make the coffee
    ordered (ol) - steps to make the coffee
    definition (dl) - some details regarding the supplies
    """
    print("testing lists")
    starbucks = Anchor(
        href="https://www.starbucks.com/menu/at-home-coffee/via-instant",
        contents="Starbucks Via instant coffee",
    )
    suppliesHeading = Heading(2, "Coffee Making Supplies")
    suppliesList = UnorderedList(
        [
            ListItem(f"coffee is from {starbucks}"),
            ListItem("knock-over resistant large coffee mug"),
            ListItem("Sugar"),
            ListItem("spoon to stir"),
            ListItem("Liquid level detector"),
            ListItem("Hot/near boiling water, from dispenser on counter."),
            ListItem("milk, real or oat, whatever is in the fridge"),
        ]
    )
    suppliesList.add_content(ListItem("testing add_content"))
    suppliesList.attr_value("aria-labelledby", suppliesHeading.id())
    oooHeading = Heading(2, "Order Of Operations")
    oooList = OrderedList(
        [
            ListItem("Collect Supplies", **{"x-test": "collect-supplies"}),
            ListItem("cut tops off coffee packets and empty contents into mug"),
            ListItem(
                "don't forget to throw empty packets in the trash, and put the scissors back where you found them"
            ),
            ListItem(
                "pour one scoop of sugar into mug, make sure no one is  watching how much sugar is used, no need to invite commentary from others"
            ),
            ListItem(
                "add hot water, amount based on listening to how full the mug sounds"
            ),
            ListItem("put level detector on rim of mug"),
            ListItem("add 'milk' until hearing high pitched 'Small World'"),
            ListItem("remove sensor from rim, wipe off probes, put back in drawer"),
            ListItem("stir"),
            ListItem(
                "microwave coffee for 1 minute, the 'milk' cooled it down too much"
            ),
            ListItem(
                "stir again after microwave.  make sure 'milk' is back in the fridge"
            ),
            ListItem("sit on stool at counter, start podcast, enjoy the coffee"),
        ]
    )
    supplies_sublist = UnorderedList(
        [
            ListItem("coffee from pantry, 3 packets.  Oh, and the sugar too"),
            ListItem("mug from drying rack"),
            ListItem("scissors from 'junk drawer'"),
        ]
    )
    oooList.add_sublist(supplies_sublist, attributes={"x-test": "collect-supplies"}, before = False)
    hidden_sublist = UnorderedList([
        ListItem("what to do when you're trying to quietly drink your coffee, and someone wants to talk to you"),
        ListItem("take your coffee to a ruum in the house that is traditionally quiet and people leave you alone"),
        ListItem("It's okay to bring your toast/muffin/biscotti as well")
    ])
    oooList.add_sublist(hidden_sublist.get_collapsible("our little secret"))
    oooList.attr_value("aria-labelledby", oooHeading.id())
    detailsHeading = Heading(2, "More Details (because I need to test the dl list)")
    detailsList = DescriptionList(
        [
            DListItem(
                [DescriptionTerm("The Coffee")],
                DescriptionDef(
                    f"I get the {starbucks} from Amazon as a subscription in the 50 count packet.  That's about as cheap as I can find it"
                ),
            ),
            DListItem(
                [DescriptionTerm("The Sugar")],
                DescriptionDef(
                    "It's, I think, the Tabago raw sugar from Trader Joe's.  it looks healthy and sounds exotic"
                ),
            ),
            DListItem(
                [DescriptionTerm("Real Milk"), DescriptionTerm("Oat Milk")],
                DescriptionDef(
                    "Why is cow's milk considered 'real'?  Or am I simply calling this out to have multiple <dt> elements for this entry?"
                ),
            ),
            DListItem(
                [DescriptionTerm("Why The list?")],
                DescriptionDef(
                    "I don't like having to use a list for dt elements, I should change that interface"
                ),
            ),
        ]
    )
    detailsList.attr_value("aria-labelledby", detailsHeading.id())
    title = "Testing Domible Lists"
    htmlDoc = basic_head_empty_body(title)
    body = htmlDoc.get_body_element()
    body.add_content(
        [
            Heading(1, "Makin' Coffee, the Coffeenator"),
            suppliesHeading,
            suppliesList,
            oooHeading,
            oooList,
            detailsHeading,
            detailsList,
        ]
    )
    open_html_in_browser(htmlDoc)


def list_builder(args) -> None:
    """ """
    title = "Testing Domible List builder"
    htmlDoc = basic_head_empty_body(title)
    body = htmlDoc.get_body_element()
    body.add_content([])
    open_html_in_browser(htmlDoc)


def headings(args) -> None:
    """
    test the creation and display of HTML headings.
    this also shows how to add attributes to elements
    Note heading level 0 and 7 do not result in legit headings to the screen reader
    For now, to verify attributes, you'll need to view source in your browser
    """
    print("testing heading generation")
    title = "Domible testing Heading Element"
    htmlDoc = basic_head_empty_body(title)
    body = htmlDoc.get_body_element()
    body.add_content(Heading(1, title))
    for lvl in range(0, 8):
        body.add_content(Heading(level=str(lvl), contents=f"Heading with Level: {lvl}"))
    body.add_content(
        Heading(
            level=1,
            contents=Anchor(
                href="https://blindgumption.com",
                contents="Blind Gumption",
                **{"class": "classValue", "id": "uniqueId"},
            ),
        )
    )
    body.add_content(
        Heading(
            className="someClass",
            level=2,
            contents="heading level 2 with someClass classname",
        )
    )
    body.add_content(
        Heading(
            className="<someClass>",
            level=2,
            contents="heading level 2 with <someClass> classname",
        )
    )
    body.add_content(
        Heading(
            uniqueId="somethingUnique",
            level=3,
            contents="heading level 3 with somethingUnique for unique",
        )
    )
    body.add_content(
        Heading(
            uniqueId="'somethingUnique'",
            level=3,
            contents="heading level 3 with 'somethingUnique' for unique",
        )
    )
    open_html_in_browser(htmlDoc)


def run() -> None:
    subparsers = global_parser.add_subparsers(
        required=True, help="sub commands for dicli"
    )

    simple_parser = subparsers.add_parser("simple", help=simple.__doc__)
    simple_parser.set_defaults(func=simple)

    elements_parser = subparsers.add_parser("elements", help=elements.__doc__)
    elements_parser.add_argument(
        "-u", "--url_mdn", default="https://developer.mozilla.org"
    )
    elements_parser.add_argument("-l", "--lang", default="en-US")
    elements_parser.add_argument("-o", "--outfile")
    elements_parser.set_defaults(func=elements)

    ctfd_parser = subparsers.add_parser("ctfd", help=ctfd.__doc__)
    ctfd_parser.add_argument("-l", "--lower", default=1, type=int, help="lower bound")
    ctfd_parser.add_argument("-u", "--upper", default=10, type=int, help="upper bound")
    ctfd_parser.set_defaults(func=ctfd)

    lists_parser = subparsers.add_parser("lists", help=lists.__doc__)
    lists_parser.set_defaults(func=lists)

    headings_parser = subparsers.add_parser("headings", help=headings.__doc__)
    headings_parser.set_defaults(func=headings)

    args = global_parser.parse_args()
    args.func(args)


## end of file



"""  domible/src/domible/elements/lists.py
Here we have elements specific to creating lists.
the list items (including <dt> and <dd>) are defined in here as well.
"""

from typing import Any

from domible.elements.baseElements import BaseElement
from domible.elements.scripting import Script, Template 
from domible.elements.inline import Details, Summary
from domible.utils import index_matching_attributes


"""
<li>, <dt>, and <dd> are nearly the same element with different tags.
<dt> has more restrictions on its permitted content but I'm not sure this package needs to enforce that.
Maybe at some point I'll implement more types around content categories and build in more enforcement.
For now, I'm taking the C approach, so go ahead and corrupt memory.
Or in this case, implement nonsensical HTML documents.
"""


class DescriptionTerm(BaseElement):
    """creates a dt element."""

    def __init__(self, contents: Any, **kwArgs):
        super().__init__(tag="dt", contents=contents, **kwArgs)


class DescriptionDef(BaseElement):
    """creates a dd element."""

    def __init__(self, contents: Any, **kwArgs):
        super().__init__(tag="dd", contents=contents, **kwArgs)


class DListItem:
    """
    This is an attempt to make the dl list element behave similar to the other list elements.
    It is designed to be analogous to an li elment.
    It is created with a list of terms and their definition.
    It renders itself since it is not derived from BaseElement.
    Screen readers seem to have trouble with dl lists, they're a pain all around.
    """

    def __init__(self, terms: list[DescriptionTerm], definition: DescriptionDef):
        self.terms = terms
        self.definition = definition

    def __repr__(self):
        entryString = ""
        for term in self.terms:
            entryString += f"{term}"
        entryString += f"{self.definition}"
        return entryString


class ListItem(BaseElement):
    """creates an li element."""

    def __init__(self, contents: Any, **kwArgs):
        super().__init__(tag="li", contents=contents, **kwArgs)


class HtmlListBase(BaseElement):
    """
    HtmlListBase is used to ensure any elements passed in at creation are either
    ListItem, Script, or Template elements 
    add_contents is overridden also to ensure valid content types 
    """
    @staticmethod
    def valid_list_element(element) -> bool:
        if element: return isinstance(element, (ListItem, DListItem, Script, Template, Details))
        return True 

    @staticmethod
    def valid_contents(contents) -> bool:
        if not isinstance(contents, list):
            return HtmlListBase.valid_list_element(contents)
        for elem in contents:
            if not HtmlListBase.valid_list_element(elem): return False 
        return True 

    def __init__(self, tag: str, entries: list[Any] = None, **kwArgs):
        if not HtmlListBase.valid_contents(entries):
            raise TypeError("list items must be either ListItem, DListItem, Script, or Template")
        super().__init__(tag, entries, **kwArgs)

    def add_content(self, content: Any, front: bool = False) -> None: 
        """
        overriding BaseElement to ensure content is appropriate for a list,
        then use BaseElement to add the contents.
        """
        if not HtmlListBase.valid_contents(content):
            raise TypeError(
                "list items must be either ListItem, DListItem, Script, or Template"
            )
        super().add_content(content, front)

    def add_item(
        self,
        item: Any,
        index: int = None,
        attributes: dict[str, str] = None,
        before: bool = True,
    ) -> None:
        """
        add an item to the list, where is based on values of other parameters.
        if index given, add item at that index
        if index is not in range, raise IndexError
        if attributes, add item based on the attributes of another item
        if before is True, ad new item before existing item with matching attributes, else after
        before is ignored if attributes is None
        if attributes is specified but no match is found, raise ValueError
        if no index and attributes, add new item to end of list
        if index and attributes are both specified, attributes is ignored.
        """
        if not HtmlListBase.valid_contents(item):
            raise TypeError(f"{item} -- is not valid list content")
        if index:
            if index >= 0 and index < len(self.contents):
                self.contents.insert(index, item)
            else:
                raise IndexError(
                    f"index {index} is outside range, len of contents is {len(self.contents)}"
                )
        elif attributes:
            if (insert := index_matching_attributes(self.contents, attributes)) != None:
                print(f"inserting at index {insert}, before is {before}")
                if before:
                    self.contents.insert(insert, item)
                else:
                    self.contents.insert(insert + 1, item)
            else:
                raise ValueError(f"no element found matching attributes {attributes}")
        else:
            self.contents.append(item)

    def add_sublist(
        self,
        sublist: BaseElement,
        index: int = None,
        attributes: dict[str, str] = None,
        before: bool = True,
    ):
        """
        add_sublist exists to address the issue of an extra bullet point
        when including an HTML list (e.g., <ul>, <ol>) within an <li>.
        This is required to comply with the content guidelines of what is allowed as a childe of a list element.
        check out the MDN docs for more details.

        The sublist must be one of the list elements from domible.elements.lists, or a details element for hidden lists.
        add_sublist will wrap it in a <li> element and
        add a style="list-style-type:none" attribute to the <li> wrapper.
        The <li> is then added to the list being built in this class/object.
        See comments for HtmlListBase.add_item for description of parameters after sublist.
        """
        li_wrapper = ListItem(sublist, style="list-style-type:none")
        self.add_item(li_wrapper, index, attributes, before)

    def get_collapsible(self, summary: any) -> Details:
        """
        I've been building up a list in self
        I want to use that list in a collapsible details/summary combination of elements 
        return the Details element with self as its hidden contents 
        """
        summary = Summary(summary)
        return Details(summary, self)


class UnorderedList(HtmlListBase):
    """creates a ul element."""

    def __init__(self, entries: list[Any] = None, **kwArgs):
        super().__init__("ul", entries, **kwArgs)


class OrderedList(HtmlListBase):
    """creates an ol element."""

    def __init__(self, entries: list[Any] = None, **kwArgs):
        super().__init__("ol", entries, **kwArgs)


class MenuList(HtmlListBase):
    """creates a menu element."""

    def __init__(self, entries: list[Any] = None, **kwArgs):
        super().__init__("menu", entries, **kwArgs)


class DescriptionList(BaseElement):
    """creates a dl element."""

    def __init__(self, entries: list[Any] = None, **kwArgs):
        super().__init__("dl", entries, **kwArgs)


## end of file



"""  domible/src/domible/elements/baseElements.py
classes in here are used to build HTML elements
most of the functionality of an element is in the base class
The derived classes are mostly responsible for setting the appropriate tag string value 
And there are places where __repr__ needs to be overwritten 

*** note about contents ***

I'll elaborate on this more in the documentation as there's possibly quite a bit to say, here's the gist of it 

Initially, I assumed contents was a single item, e.g., a string, or another BaseElement.
I realized some elements had lists of other elements as their contents, so I made another class, BaseElementList 
in which the contents was specifically a list, though the list could hold anything.
As I learned more about HTML and was developing more complex documents, 
I realized I should only have BaseElement and let contents be anything, including a list.
I could handle it all within the BaseElement implementation. 
Thus BaseElementList is gone and BaseEllement dynamically adjust contents accordingly. 
"""

from __future__ import annotations

from random import random 
from typing import Any

from domible.utils import is_subdict 


class ContentsList(list):
    """
    this class only exists because list.__repr__ prints the square brackets 
    when it's generating the HTML for the list objects  
    """
    def __repr__(self):
        """ 
        return a string that does not include square brackets for the list 
        and has a newline between each element, instead of comma 
        This is used to generate HTML text for sibling elements 
        """
        #  I should change this to map(), except this way is so readable ... 
        listString = '\n'
        for entry in self:
            listString += f'{entry}\n'
        return listString 


class BaseElement:
    """
    BaseElement holds the tag value, the contents,  and a dict of attrivutes      
    and some methods likely needed by all elements 

    The text for the element is genereated by __repr__
    and almost all elements derived from BaseElement can let BaseElement.__repr__ generate the text

    Should we deep copy the contents and attributes?
    Probably not the attributes, they're all strings.
    The contents though can be very complicated structures 
    with many levels of nesting of other elements.
    If someone used elementX in the contents of elementB, 
    then changed elementX, that would change elementB if only references were copied.

    Should it then be an explicit callout that contents are references, not deep copied?
    Maybe ...    But isn't that just saying Python uses pass by reference?
    """
    def __init__(self, tag: str, contents: Any = None, **kwArgs):
        self.tag = tag
        if not contents:
            self.contents = ContentsList()
        elif not isinstance(contents, list):
            # need to call this out here so we don't end up with a list of a list 
            self.contents = ContentsList([contents])
        else:
            self.contents = ContentsList(contents)
        self.attributes = dict(kwArgs)

    def get_elements(self, tag: str = None, **attrs) -> list[Any]:
        """
        look at this element and any elements in its contents and 
        return any elements with the given tag and list of attributes 
        Search is breadth first with returned elements in a list 
        with an ordering you'd expect from BFS 

        Think of self as the root of a generic tree.
        self has contents which could be empty, have one or more BaseElements, or something else (e.g., string).
        Only other BaseElements are considerred nodes needing to be searched.
        Nodes not derived from BaseElement are considered  leave nodes, and not searched. 
        """
        found = []
        if not (tag or attrs): return found
        searching = [self]
        while len(searching) > 0:
            current = searching.pop(0)
            if tag == current.tag and is_subdict(attrs, current.attributes):
                found.append(current)  
            searching += [elem for elem in current.contents if isinstance(elem, BaseElement)]    
        return found 

    def attr_value(self, attr: str, value: str = None) -> str:
        """
        use this to set a value for a single attribute,
        or to get the value set for the given attribute.
        if value is None, returns the value of attr in self.attrs
        if value is set, adds attr to attrs with given value.
        if attr already existed in attrs, updates attr in attrs with new value and  returns old value 
        """ 
        originalValue = self.attributes.get(attr)
        if value:
            self.attributes[attr] = value
            # if attr did not exist already, need to return value it was just set to
            # if originalValue is not None, attr already existed thus need to return originalValue  
            originalValue = value if not originalValue else originalValue 
        return originalValue 

    def id(self, value:str = None) -> str:
        """
        id() is separate from attrValue as it's reasonable for the package to generate a unique id for the element.
        If there is no id in self.attrs, id() will generate one, add it to attrs and return the value.
        if the value argument is provided, id() will go through attrValue to set that value for the 'id' attribute 
        and return whatever attrValue returned.
        """
        if value:
            return self.attr_value('id', value)
        elif (existing := self.attributes.get('id')):
            # no value was provided and 'id' already exists in self.attyrs, return the existing value 
            return existing
        else:
            # no value was provided and 'id' does not already exist in self.attrs
            # generate an id, set it in attrs and return that new value 
            idValue = f'{self.tag}-{random()}'
            self.attr_value('id', idValue)
            return idValue 

    def get_attributes_string(self, inAttrs: dict = None) -> str:
        """
        in general, this will return the string of the attributes currently set on the element
        alternatively, if a derived element has a dict of attributes not in the base attrs dict,
        that dict can be passed in and used instead.
        """
        lattrs = inAttrs if inAttrs else  self.attributes 
        attrString = ''
        if lattrs and len(lattrs) > 0: 
            for k,v in lattrs.items():
                attrString += f' {k}="{v}"'
        return attrString 


    def add_attributes(self, **kwArgs) -> None:
        """ add all atttributes in kwArgs to self.attrs """
        self.attributes.update(kwArgs)


    def add_content(self, content: Any, front: bool = False) -> None: 
        """ 
        add content to any existing contents.  
        Adds to end of contents by default 
        We know from __init__, contents is at least an empty list 
        """
        if not isinstance(content, list):
            content = [content]
        if front:
            self.contents = ContentsList(content + self.contents)
        else:
            self.contents = ContentsList(self.contents + content)

    def set_content(self, content: Any) -> None: 
        """ 
        set the content of this element to the passed in content
        any existing content will be be lost 
        """
        if not isinstance(content, list):
            content = [content]
        self.contents = ContentsList(content)


    def opening_tag(self) -> str:
        """ returns a string representing the opening tag  including any attributes """
        return f'<{self.tag}{self.get_attributes_string()}>'
        
    def closing_tag(self) -> str:
        """ returns closing tag """
        return f'</{self.tag}>'


    def __repr__(self):
        return f'{self.opening_tag()}{self.contents}{self.closing_tag()}'


class BaseVoidElement(BaseElement):
    """
    a void element cannot have contents and there is no closing tag
    This class enforces those two properties by overriding a few of BaseElements methods 
    """    
    def __init__(self, tag: str, **kwArgs):
        super().__init__(tag=tag, **kwArgs)

    def add_content(self, content: Any, front: bool = False) -> None: 
        """ 
        void elements do not contain content 
        as it is now, simply return from this method.
        Maybe we should raise an exception 
        or provide an override to say do it regardless.
        If an override is eventually added, __repr__ will need to be updated to include the contents 
        """
        return 

    def __repr__(self):
        """ a void element only has an opening tag, with attributes, if any  """
        return f'{self.opening_tag()}'


## end of file 


""" domible/src/distarter/main.py 

this file serves as Python code to get started with domible
"""

from domible.elements import Html, Body, Title
from domible.elements import Heading, Anchor, Paragraph, HorizontalRule 
from domible.starterDocuments import basic_head_empty_body
from domible.tools import open_html_in_browser


#######
def run() -> None:
    """
    generate, and show in browser,
    a very simple HTML document
    """
    title = "BareBones with a Minimal Body, though getting less minimal as I use this for testing..."
    cib_anchor = Anchor(
        href="http://codinginblind.vip", contents="the Coding In Blind website"
    )
    htmlDoc = basic_head_empty_body(title)
    body = htmlDoc.get_body_element()
    body.add_content([
        Heading(1, title), 
        Paragraph(f"check out: {cib_anchor}"),
        HorizontalRule(),
        Paragraph('Adding a non printable character, codepoint = 0xfeff, in the square brackets [\ufeff] to make sure encodings are set correctly.')
    ])
    open_html_in_browser(htmlDoc)
    ## open_html_in_browser(htmlDoc, "saved_distarter.html", True)


if __name__ == "__main__":
    run()

## end of file



""" domible/src/domible/elements/roots.py 

here is the <html> tag, considered the root of an HTML document.
"""

import logging

logger = logging.getLogger(__name__)

from typing import Any

from domible.elements.baseElements import BaseElement
from domible.elements.meta import Head
from domible.elements  .sectioning import Body


class Html(BaseElement):
    """ 
    creates the root <html> element
    when rendered, also renders <!DOCTYPE html> 
    before the opening <html> tag 
    """
    def __init__(self, lang: str = "en", contents: Any = None, **kwArgs):
        self.lang = lang
        super().__init__(tag="html", contents=contents, lang=lang, **kwArgs)

    def get_head_element(self) -> Head:
        """
        return the Head element from the html doc
        Only return it if it's the only one and is an instance of Head.
        else log a warning and return None
        """
        elms = self.get_elements("head")
        if len(elms) == 1 and isinstance(elms[0], Head):
            # make this the first check as it's almost certainly true
            return elms[0]
        elif len(elms) == 0:
            logger.warning("html doc has no Head!  good pour!")
            return None
        elif len(elms) > 1:
            logger.warning(
                "html doc appears to have more than one Headelement, Cerberus?!?"
            )
            return None
        else:
            #  not isinstance(elms[0], Head):
            logger.warning(
                "html doc has an element with tag head, but element is not of type Head.  I blame the user!"
            )
            return None

    def get_body_element(self) -> Body:
        """
        return the body element from the html doc
        Only return it if it's the only one and is an instance of Body.
        else log a warning and return None
        """
        elms = self.get_elements("body")
        if len(elms) == 1 and isinstance(elms[0], Body):
            # make this the first check as it's almost certainly true
            return elms[0]
        elif len(elms) == 0:
            logger.warning("html doc has no body!  By the Geeky Blues Band")
            return None
        elif len(elms) > 1:
            logger.warning(
                "html doc appears to have more than one body element, BAD Doc!!"
            )
            return None
        else:
            #  not isinstance(elms[0], Body):
            logger.warning(
                "html doc has an element with tag body, but element is not of type Body.  I blame the user!"
            )
            return None

    def __repr__(self):
        return f"<!DOCTYPE html> \n{super().__repr__()}"


# end of file



"""  domible/src/domible/elements/inline.py 
"""
from typing import Dict, Any

from domible.elements.baseElements import BaseElement


class Anchor(BaseElement):
    def __init__(self, href: str, contents: Any = None, **kwArgs):
        self.href = href
        super().__init__(tag="a", contents=contents, href=self.href, **kwArgs)


class Div(BaseElement):
    """create a div element"""

    def __init__(self, contents: Any = None, **kwArgs):
        super().__init__(tag="div", contents=contents, **kwArgs)


#######
class Paragraph(BaseElement):
    def __init__(self, contents: Any = None, **kwArgs):
        super().__init__(tag="p", contents=contents, **kwArgs)


#######
class Span(BaseElement):
    """create a span element
    use for wrapping a single content .
    For multiple elements, use a div.
    """

    def __init__(self, contents: Any = None, **kwArgs):
        super().__init__(tag="span", contents=contents, **kwArgs)


class Summary(BaseElement):
    """create a Summary element
    used within a details element to provide a label for a collapseable widget.
    For accessibility, requiring some content be provided
    """

    def __init__(self, contents: Any, **kwArgs):
        if not contents:
            raise ValueError("Summary element must have some contents")
        super().__init__(tag="summary", contents=contents, **kwArgs)


class Details(BaseElement):
    """create a details element
    used to create a collapseable widget.
    For accessibility, requiring a summary element be provided
    """

    def __init__(self, summary: Summary,  contents: Any = None, **kwArgs):
        if not summary or not isinstance(summary, Summary):
            raise ValueError("Details element must contain a Summary element")
        if isinstance(contents, list): 
            contents = contents.insert(0, summary)
        else:
            contents  = [summary] + [contents] if contents else summary
        super().__init__(tag="details", contents=contents, **kwArgs)


## end of file



""" /domible/src/dicli/mdnElements.py
code to get HTML elements reference from MDN
used by getTableData 

this module is very specific to the layout of the MDN HTML reference pages

As the MDN site is scraped, TableInfo objects are constructed with a row for each element.
A table for current elements and one for deprecated is constructed.

Yes, this has devolved into much hackery.
I should turn this in to a class and use properties
and get rid of the globals.
The globals were initially hard coded "const" values.
When mdn_base_url and lang (language) became parameters, the const became globals
Someday... I'll fix this??  
"""

import logging

logger = logging.getLogger(__name__)

from os import path
from random import random 
from time import sleep 

from bs4 import BeautifulSoup as BSoup
from bs4.element import Tag
import requests
from typing import Dict, Tuple

from domible.elements import Anchor
from domible.builders.tableBuilder import TableBuilder, RowBuilder


MdnLocalPathPrefix = "" 
ElementsReferencePath = ""
MdnHostUrlBase = ""
ElementsReferenceUrl = f"{MdnHostUrlBase}{ElementsReferencePath}"
MdnAnchor = None 

DomibleIssuesPage = Anchor(href="https://github.com/joeldodson/domible/issues", contents="domible issue page on github")

def updateInPageAnchorHref(soup: BSoup, elemUrl: str) -> None:
    """
    any anchors with href to local id, 
    e.g., href="#somewherein page"
    need to be updated to an absolute (because absolute is the url I have) href
    """ 
    anchors = soup.find_all("a")
    for a in anchors:
        href = a.get("href")
        if href and href.startswith("#"):
            a["href"] = f"{elemUrl}{href}"


def filterAnchor(anchor: Tag) -> bool:
    """
    for each element, there is a page with a url of the form
    ElementsReferenceUrl/<element_name>
    there are other anchors on the base MDN HTML reference page though with a similar base path
    there is also an anchor with href that looks like its fro an element called 'contributors.txt', filter that out too
    This filter is to ensure we get only the anchors referencing HTML elements
    """
    href = anchor.attrs.get("href")
    if not href or ElementsReferencePath not in href or "contributors.txt" in href:
        return False
    elif path.split(href)[0] != ElementsReferencePath:
        return False
    return True


def getElementSummary(soup: BSoup) -> Dict:
    """
    find and return the summary in the document,

    NOTE: this method hard codes knowledge of the format of the document.
    consider the first paragraph after the main element opening tag,
    and all its sibling paragraphs, to be the summary.
    """
    summary = ""
    try:
        p0 = soup.find("main").find("p")
        summary = p0.decode()
        for el in p0.next_siblings:
            if el.name and el.name == 'p': summary += el.decode()
            elif el.name: break
    except Exception:
        logger.exception("exception while getting summary")
        summary = f"no summary?  that's odd.  Please open an issue at the {DomibleIssuesPage}"
    return {"Summary": summary}


def getElementSpecificationReference(soup: BSoup) -> Dict:
    """
    The element specification is in a table with a single column and two rows.
    The first row is a <thead> with a <th> with the word "Specification"
    The second row is the specification link in a <td> in a <tbody>
    This is generally the second table on the page, but let's not assume that 
    """
    entry = dict()
    for table in soup.find_all('table'):
        if (thead := table.find('thead')) and thead.text.lower() == "specification":
            entry["Specification"] = table.find('tbody').find('td').decode_contents()
    return entry 


thToUseList = []
def findThToUse(th: Tag) -> Tag:
    """
    see notes on getElementTechnicalSummary for why this is here
    And the global hack is the easiest way to fake C like static variables 
    """
    global thToUseList
    for savedTh in thToUseList:
        if savedTh.text.lower() == th.text.lower():
            return savedTh
    thToUseList.append(th)
    return th


def getElementTechnicalSummary(soup: BSoup) -> Dict:
    """
    The page for each element has a two column table with summaries of various topics.
    The first column is the topic, the second is the summary.
    I want the topics to be columns in the table we're creating, the topic summary will be in the row for each element.
    Unfortunately, not all the tables topic columns are consistent.
    If I simply pull out the content of the th, and use that as a key in the dict for the row,
    we end up with multiple columns for what should be the same column.
    For example, "DOM Interface" and "DOM interface" turn into two different columns.
    To address this, the entries from the first element, <a> are saved to be used as keys for all the elements.
    Subsequent element table entries need to be compared against the saved keys. 
    """
    entries = dict() 
    detailstable = soup.find("table", {"class":"properties"})
    if detailstable:
        ths = detailstable.find_all('th')
        tds = detailstable.find_all('td')
        for th,td in zip(ths, tds):
            thToUse = findThToUse(th)
            entries[thToUse.decode_contents()] = td.decode_contents()
    return entries 


def getElementDetails(soup: BSoup, elemUrl: str) -> Dict:
    """
    get all the information from the element's html page we want for the table
    A high level summary is in the first <p> element immediately after the only <h1> element in the doc
    A technical summary is in the first table.
    we'll use the first column as column headers for the table we're building.
    The second column is the values
    The second table in the doc has a link to the specification for the element.
    """
    result = dict()
    if d := getElementSummary(soup):
        result.update(d)
    if d := getElementSpecificationReference(soup):
        result.update(d)
    if d := getElementTechnicalSummary(soup):
        result.update(d)
    return result 


def getElementInformation(elemName: str, elemUrl: str) -> Dict:
    """
    request.get the page for the element
    then parse it to find what we're looking for
    """
    try:
        elemDoc = requests.get(elemUrl)
        elemSoup = BSoup(elemDoc.content, "html.parser")
        updateInPageAnchorHref(elemSoup, elemUrl)
        return getElementDetails(elemSoup, elemUrl)
    except Exception:
        logger.exception(f"failed while getting info for element {elemName}")
        logger.fatal(f"failed while getting info for element {elemName}")
        exit(1)


def getElementsTables(mdnBaseUrl: str, lang: str) -> Tuple[TableBuilder, TableBuilder]:
    global MdnLocalPathPrefix, ElementsReferencePath, MdnHostUrlBase, ElementsReferenceUrl, MdnAnchor 
    MdnLocalPathPrefix = f"/{lang}/docs" ## was  "/en-US/docs/Web"
    ElementsReferencePath = f"{MdnLocalPathPrefix}/Web/HTML/Element"  ## was "/en-US/docs/Web/HTML/Element"
    MdnHostUrlBase = mdnBaseUrl  ## was "https://developer.mozilla.org"
    ElementsReferenceUrl = f"{MdnHostUrlBase}{ElementsReferencePath}"
    MdnAnchor = Anchor(
        ElementsReferenceUrl,
        "HTML elements reference page on Mozilla Developer Network (MDN)",
    )

    try:
        refPage = requests.get(ElementsReferenceUrl)
        soup = BSoup(refPage.content, "html.parser")
        anchors = soup.find_all("a")
        elems = list(
            set([a.attrs.get("href").lower() for a in anchors if filterAnchor(a)])
        )
        elems.sort()
        logger.info(f"number of element references is {len(elems)}")
        # we have links to each element,
        # time to scrape each page and create rows for each element
        currentElementsTable = TableBuilder(
            caption="Current HTML Elements", row_heading_name="Element"
        )
        deprecatedElementsTable = TableBuilder(
            caption="Deprecated HTML Elements", row_heading_name="Element"
        )
        for elmPath in elems:
            sleep(random())  # don't get blocked by MDN...
            elemUrl = f"{MdnHostUrlBase}{elmPath}"
            elemName = path.split(elemUrl)[1]
            logger.info(f"getting element {elemName} from path {elemUrl}")
            rowEntries = getElementInformation(elemName, elemUrl)
            row = RowBuilder(Anchor(elemUrl, elemName), rowEntries)
            if "Deprecated" in rowEntries.get("Summary"):
                deprecatedElementsTable.add_row(row)
            else:
                currentElementsTable.add_row(row)
    except Exception:
        logger.exception("something bad happened while scraping MDN")
        return (None, None)
    return (currentElementsTable, deprecatedElementsTable)


## end of file



""" domible/src/domible/tools.py

tools.py has functions for users of the domible package to do simple things
like open their HTML doc in the browser,
or save the HTML to a specified file.
tools.py imports other modules within domible thus should not be used by any of domible's submodules.
Any helper functions needed by domible should be put in utils.py 
See the top level comments in utils.py to understand why tools.py exists. 
"""

from pathlib import Path 
from tempfile import NamedTemporaryFile
import webbrowser as wb

from domible.elements import BaseElement 
from domible.elements import Html, Body 
from domible.builders import element_from_object 
from domible.starterDocuments import basic_head_empty_body 


def save_to_file(element: BaseElement, filename: str, force: bool = False) -> None:
    """
    save the passed in element to passed in filename.
    If the file name exists and is a regular file,
    save will fail unless force is Tru
    """
    fp = Path(filename)
    if fp.exists():
        # if the file exists, is a regular file and force is true, carry on
        # otherwise, raise a FileExists error
        if not fp.is_file() or not force:
            raise FileExistsError(
                f"{filename} exists and is not a regular file, or force is False"
            )
    # if file does exists, force must be True
    with fp.open("w+t", encoding="utf-8") as f:
        f.write(f"{element}")


def open_html_in_browser(html_doc: Html, save_file: str = None, force: bool = False) -> None:
    """
    open the html_doc in the default browser.
    if a save_to_file is provided, also save the html_doc to that file,
    else use a temporary file 
    """
    if save_file:
        save_to_file(html_doc, save_file, force)
        # path must be absolute to match how temp file works 
        path = str(Path(save_file).absolute()) 
    else: 
        path =NamedTemporaryFile(delete=False, suffix='.html')
        f=open(path.name, 'w+t', encoding='utf-8')
        f.write(f"{html_doc}")
        f.close()
        path = path.name # consistent with path from saving file 
    wb.open('file://' + path)


def open_object_in_browser(obj: object, depth:int = 42, title: str = "opening an object in the browser", save_file: str = None, force: bool = False) -> None:
    """
    get HTML representation of the object then open it in the default browser.

    the passed in object is used to get HTML using element_from_object.
    That HTML is then put into a simple HtML document using the basicHeadEmptyBody starter document.
    that simple document is then opened in the browser. 
    """
    obj_html = element_from_object(obj, depth)
    html_doc: Html  = basic_head_empty_body(title)
    body: Body = html_doc.get_body_element() 
    body.add_content(obj_html)
    open_html_in_browser(html_doc, save_file, force)


## end of file



""" domible/src/domible/utils.py 
little things I need to keep the code clean.

I'm writing the following note after creating the tools.py module at the same level as utils.py.
Hopefully this clarifies why there is a utils.py and tools.py, maybe it's just my bad planning...
utils.py is imported in many of the submodules in domible
thus should not import (depend on) anything within domible.
It shuld mainly be generic Python code to keep domible code cleaner.
Nothing in utils is intended to be used by users of the domible package.
See comments in tools.py for more clarification. 
"""

from typing import Any 


def is_subdict(dSub: dict[str,str], dSuper: dict[str,str]) -> bool:
    """
    if all attributes in dSub are also in dSuper,
    and all values match,
    return True
    else False

    NOTE: treat None as empty set 
    """
    dSub = dSub if dSub  else {}
    dSuper = dSuper if dSuper else {} 
    subKV = {(k,v) for k,v in dSub.items()}
    superKV = {(k,v) for k,v in dSuper.items()}
    return subKV.issubset(superKV)


def index_matching_attributes(items: list[Any], attributes: dict[str,str]) -> int:
    """
    given a list of elements and a dict of attributes,
    return the index of the first element in the list for which attributes is a subset of its attributes 
    return None if no element has matching attributes 
    """
    for index, item in enumerate(items):
        if is_subdict(attributes, item.attributes):
            print(f"indexMatchinAttributes returning index: {index}")
            return index
    return None 


## end of file



""" domible/src/domible/builders/__init__.py """

from domible.builders.tableBuilder import RowBuilder, TableBuilder, build_table_from_dicts
from domible.builders.navBuilder import NavBuilder 
from domible.builders.elementFromObject import element_from_object  

# end of file
